/*
 *  RTEMS/FreeBSD driver for NS8390 Ethernet Chipset
 *
 *  This is a generic NS8390 network driver and requires a mini-driver
 *  to provide board/system specific functionality.
 *
 *  Author:
 *    David Fiddes, D.J@fiddes.surfaid.org
 *    http://www.calm.hw.ac.uk/davidf/coldfire/
 *
 *  COPYRIGHT (c) 1989-1998.
 *  On-Line Applications Research Corporation (OAR).
 *  Copyright assigned to U.S. Government, 1994.
 *
 *  The license and distribution terms for this file may be
 *  found in the file LICENSE in this distribution or at
 * 
 *  http://www.OARcorp.com/rtems/license.html.
 * 
 *  $Id: ns8390.c,v 1.11 1998/02/19 23:02:02 joel Exp $
 */

#include <bsp.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdarg.h>
#include <rtems/error.h>
#include <rtems/rtems_bsdnet.h>

#include <sys/param.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/sockio.h>

#include <net/if.h>

#include <netinet/in.h>
#include <netinet/if_ether.h>

#include <ns8390.h>


static const char *version = 
  "NS8390 RTEMS/FreeBSD sub-driver A0.01 22/08/98 David Fiddes (D.J@fiddes.surfaid.org)\n";

/* 0 = no debug, 1 = standard traces, 2 = register accesses, 3 = all accesses inc data */

#define DEBUG_NONE    0
#define DEBUG_STD     1
#define DEBUG_REG     2
#define DEBUG_DATA    3

static const int debug_level = DEBUG_NONE;


/*
 * Number of NS8390s supported by this driver
 */
#define NUM_NS8390_DEVICES  1 /* we'll keep things very simple for now */

/*
 * RTEMS event used by interrupt handler to signal driver tasks.
 * This must not be any of the events used by the network task synchronization.
 */
#define INTERRUPT_EVENT RTEMS_EVENT_1

/*
 * RTEMS event used to start transmit daemon.
 * This must not be the same as INTERRUPT_EVENT.
 */
#define START_TRANSMIT_EVENT  RTEMS_EVENT_2

/*
 * Receive buffer size -- Allow for a full ethernet packet including CRC
 */
#define RBUF_SIZE 1520

#if (MCLBYTES < RBUF_SIZE)
# error "Driver must have MCLBYTES > RBUF_SIZE"
#endif

/*
 * Basic Ethernet constants
 */
#define ET_MINLEN 60

/* Interrupts we want to handle from the device.  */

#define NS8390_INTERRUPTS \
  (NS8390_ISR_PRX | NS8390_ISR_PTX | NS8390_ISR_RXE | NS8390_ISR_TXE | NS8390_ISR_OVW | NS8390_ISR_CNT)


/*
 * Hardware-specific storage
 */

static struct ns8390_device ns8390_devices[NUM_NS8390_DEVICES]; 


/*
 * Module prototypes
 */
static void
ns8390_stop (struct ns8390_device *dp);

void
ns8390_tx_daemon (void *arg);

static void
ns8390_sendpacket (struct ifnet *ifp, struct mbuf *m);

static void
ns8390_rx_daemon (void *arg);

static rtems_isr
ns8390_interrupt_handler (rtems_vector_number v);

static void
ns8390_start (struct ifnet *ifp);


static void
ns8390_init (void *arg);

static void
ns8390_stats (struct ns8390_device *dp);

void
ns8390_initialise( struct ns8390_device *dp );

void
ns8390_startup( struct ns8390_device *dp );



/* 
 * Find the NS8390 device which is attached at a particular interrupt
 * vector.
 */
static struct ns8390_device *
ns8390_device_for_irno (int irno)
{
  int i;

  for (i = 0; i < NUM_NS8390_DEVICES; ++i)
  {
    if ( ns8390_devices[i].irno == irno &&
         ns8390_devices[i].arpcom.ac_if.if_softc != NULL )
      return &ns8390_devices[i];
  }

  return NULL;
}


/* 
 * Handle the current NS8390 status.  This is called when the device
 * signals an interrupt.  It is also called at other times while
 * NS8390 interrupts have been disabled.
 */

static void
ns8390_check_status (struct ns8390_device *dp)
{
  unsigned32    ns8390_port;

  setRegister_f setReg;
  getRegister_f getReg;
  getData_f     getData;
  
  unsigned char status;

  /* setup the IO functions */
  ns8390_port = dp->port;
  setReg      = dp->setRegister;
  getReg      = dp->getRegister;
  getData     = dp->getData;


  /* It seems that we need to use a loop here, because if the NE2000
     signals an interrupt because packet transmission is complete, and
     then receives a packet while interrupts are disabled, it seems to
     sometimes fail to signal the interrupt for the received packet
     when interrupts are reenabled.  (Based on the behaviour of the
     Realtek 8019AS chip).  */
  /* TODO: Determine the heritage of this statement - DJF */

  while (1) {
    status = (*getReg)( ns8390_port, NS8390_PAGE0_READ_ISR );
      
    if (status == 0)
      break;

#ifdef DEBUG_NE2000
    printk ("NE2000 status 0x%x (8259 enabled: %s; mask: %x)\n", status,
	    i8259s_cache & (1 << sc->irno) ? "no" : "yes",
	    i8259s_cache);
#endif

    /* Check for incoming packet overwrite.  */
    if (status & NS8390_ISR_OVW) {
      unsigned char status2;

      ++dp->eth_stats.rx_over_errors;
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB | NS8390_CR_STP );
      Wait_X_ms (2);
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_RBCR0, 0 );
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_RBCR1, 0 );
      status2 = (*getReg)( ns8390_port, NS8390_PAGE0_READ_ISR );
      status |= status2 & (NS8390_ISR_PTX | NS8390_ISR_TXE);
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_TCR, NS8390_TCR_LOOP_I0 );
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB | NS8390_CR_STA );
      dp->overrun = 1;
      if ((status & (NS8390_ISR_PTX | NS8390_ISR_TXE)) == 0)
	dp->resend = 1;
      rtems_event_send (dp->rx_daemon_tid, INTERRUPT_EVENT);
    }

    /* Check for transmitted packet.  The transmit daemon may now be
       able to send another packet to the device.  */
    if ((status & (NS8390_ISR_PTX | NS8390_ISR_TXE)) != 0) {
      ++dp->eth_stats.tx_acks;
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, status & (NS8390_ISR_PTX | NS8390_ISR_TXE) );
      --dp->inuse;
      dp->transmitting = 0;
      if (dp->inuse > 0 || (dp->arpcom.ac_if.if_flags & IFF_OACTIVE) != 0)
	rtems_event_send (dp->tx_daemon_tid, START_TRANSMIT_EVENT);
    }

    /* Check for received packet.  */
    if ((status & (NS8390_ISR_PTX | NS8390_ISR_TXE)) != 0) {
      ++dp->eth_stats.rx_acks;
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, status & (NS8390_ISR_PTX | NS8390_ISR_RXE) );
      rtems_event_send (dp->rx_daemon_tid, INTERRUPT_EVENT);
    }

    /* Check for counter change.  */
    if ((status & NS8390_ISR_CNT) != 0) {

      dp->eth_stats.rx_frame_errors  += (*getReg)( ns8390_port, NS8390_PAGE0_READ_CNTR0 );
      dp->eth_stats.rx_crc_errors    += (*getReg)( ns8390_port, NS8390_PAGE0_READ_CNTR1 );
      dp->eth_stats.rx_missed_errors += (*getReg)( ns8390_port, NS8390_PAGE0_READ_CNTR2 );
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_CNT );
    }
  }

  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB | NS8390_CR_STA );
}


/* Handle an NS8390 interrupt.  */

static rtems_isr
ns8390_interrupt_handler (rtems_vector_number v)
{
  struct ns8390_device *dp;

  dp = ns8390_device_for_irno (v);
  if (dp == NULL)
    return;

  ++dp->eth_stats.interrupts;

  ns8390_check_status (dp);
}


/*
 * Device functions
 */

/* Stop an NS8390 */

static void
ns8390_stop (struct ns8390_device *dp)
{
  unsigned32    ns8390_port;

  setRegister_f setReg;
  getRegister_f getReg;
  getData_f     getData;

  int           i;

  if (debug_level >= DEBUG_STD)
    printf("ns8390_stop() - start\n");

  /* setup the IO functions */
  ns8390_port = dp->port;
  setReg      = dp->setRegister;
  getReg      = dp->getRegister;
  getData     = dp->getData;

  dp->arpcom.ac_if.if_flags &= ~IFF_RUNNING;

  /* Stop everything.  */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_RD_AB | NS8390_CR_STP );

  /* Wait for the interface to stop, using I as a time limit.  */
  for (i = 0; i < 5000; ++i)
    {
      unsigned char status;

      status = (*getReg)( ns8390_port, NS8390_PAGE0_READ_ISR );
      if ((status & NS8390_ISR_RST) != 0)
	break;
    }

  dp->inuse = 0;
  dp->nextavail = 0;
  dp->nextsend = 0;
  dp->transmitting = 0;

  if (debug_level >= DEBUG_STD)
    printf("ns8390_stop() - end\n");

}


/*
 * Driver transmit daemon
 */
#if 0
void
ns8390_tx_daemon (void *arg)
{
  struct ns8390_device  *dp = (struct ns8390_device *)arg;
  struct ifnet      *ifp = &dp->arpcom.ac_if;
  struct mbuf       *m;
  rtems_event_set     events;

  if (debug_level >= DEBUG_STD)
    printf( "ns8390_tx_daemon() - start\n" );

  for (;;) {
    
    if (debug_level >= DEBUG_STD)
      printf( "Waiting to TX a packet...\n" );

    /*
     * Wait for packet
     */
    rtems_bsdnet_event_receive (START_TRANSMIT_EVENT, RTEMS_EVENT_ANY | RTEMS_WAIT, RTEMS_NO_TIMEOUT, &events);

    /*
     * Get the next mbuf chain to transmit.
     */
    IF_DEQUEUE(&ifp->if_snd, m);
    if ( m != NULL)
      ns8390_sendpacket (ifp, m);

    ifp->if_flags &= ~IFF_OACTIVE;
  }
}

#endif

/* Load an NE2000 packet onto the device.  */

static void
ne_loadpacket (struct ne_softc *sc, struct mbuf *m)
{
  unsigned int port = sc->port;
  unsigned int dport = port + DATAPORT;
  struct mbuf *mhold = m;
  int leftover;
  unsigned char leftover_data;
  int timeout;

#ifdef DEBUG_NE2000
  printk ("Uploading NE2000 packet\n");
#endif

  /* Reset remote DMA complete flag.  */
  outport_byte (port + ISR, MSK_RDC);

  /* Write out the count.  */
  outport_byte (port + RBCR0, m->m_pkthdr.len);
  outport_byte (port + RBCR1, m->m_pkthdr.len >> 8);

  sc->sendlen[sc->nextavail] = m->m_pkthdr.len;

  /* Tell the device which address we want to write to.  */
  outport_byte (port + RSAR0, 0);
  outport_byte (port + RSAR1,
		NE_FIRST_TX_PAGE + (sc->nextavail * NE_TX_PAGES));

  /* Set up the write.  */
  outport_byte (port + CMDR, MSK_PG0 | MSK_RWR | MSK_STA);

  /* Transfer the mbuf chain to device memory.  NE2000 devices require
     that the data be transferred as words, so we need to handle odd
     length mbufs.  */

  leftover = 0;
  leftover_data = '\0';

  for (; m != NULL; m = m->m_next) {
    int len;
    unsigned char *data;

    len = m->m_len;
    if (len == 0)
      continue;

    data = mtod (m, unsigned char *);

    if (leftover) {
      unsigned char next;

      /* Data left over from previous mbuf in chain.  */
      next = *data++;
      --len;
      outport_word (dport, leftover_data | (next << 8));
      leftover = 0;
    }

    while (len > 1) {
      outport_word (dport, data[0] | (data[1] << 8));
      data += 2;
      len -= 2;
    }

    if (len > 0)
      {
	leftover = 1;
	leftover_data = *data++;
      }
  }

  if (leftover)
    outport_word (dport, leftover_data);

  m_freem (mhold);

  /* Wait for the device to complete accepting the data, with a
     limiting counter so that we don't wait too long.  */
  for (timeout = 0; timeout < 100; ++timeout)
    {
      unsigned char status;

      inport_byte (port + ISR, status);

#ifdef DEBUG_NE2000
      if ((status &~ MSK_RDC) != 0)
	printk ("Status 0x%x while waiting for acknowledgement of uploaded packet\n",
		status);
#endif

      if ((status & MSK_RDC) != 0) {
	outport_byte (port + ISR, MSK_RDC);
	break;
      }
    }

  if (timeout >= 100)
    printk ("Timed out waiting for acknowledgement of uploaded NE2000 packet\n");

  ++sc->nextavail;
  if (sc->nextavail == NE_TX_BUFS)
    sc->nextavail = 0;
}

/* Tell the NS8390 to transmit a buffer whose contents we have already
   loaded onto the device.  */

static void
ns8390_transmit (struct ns8390_device *dp)
{
  int len;

  unsigned32            ns8390_port;

  setRegister_f         setReg;
  getRegister_f         getReg;
  getData_f             getData;

  /* setup the IO functions */
  ns8390_port = dp->port;
  setReg      = dp->setRegister;
  getReg      = dp->getRegister;
  getData     = dp->getData;

#ifdef DEBUG_NE2000
  printk ("Transmitting NE2000 packet\n");
#endif

  len = dp->sendlen[dp->nextsend];
  if (len < ET_MINLEN)
    len = ET_MINLEN;

  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_TBCR0, len & 0xFF );
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_TBCR1, (len >> 8) & 0xFF );

  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_TPSR, NS8390_FIRST_TX_PAGE + (dp->nextsend * NS8390_TX_PAGES) );

  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB | NS8390_CR_TXP | NS8390_CR_STA );

  ++dp->nextsend;
  if (dp->nextsend == NS8390_TX_BUFS)
    dp->nextsend = 0;

  ++dp->eth_stats.tx_packets;
}


/* The NS8390 packet transmit daemon.  This task is started when the
   NS8390 driver is initialized.  */

void
ns8390_tx_daemon (void *arg)
{
  struct ns8390_device  *dp = (struct ns8390_device *)arg;
  struct ifnet          *ifp = &dp->arpcom.ac_if;

  unsigned32            ns8390_port;

  setRegister_f         setReg;
  getRegister_f         getReg;
  getData_f             getData;

  /* setup the IO functions */
  ns8390_port = dp->port;
  setReg      = dp->setRegister;
  getReg      = dp->getRegister;
  getData     = dp->getData;

  if (debug_level >= DEBUG_STD)
    printf( "ns8390_tx_daemon() - start\n" );

  while (1) {
    rtems_event_set events;

    /* Wait for a packet to be ready for sending, or for there to be
       room for another packet in the device memory.  */
    rtems_bsdnet_event_receive ( START_TRANSMIT_EVENT,
                                 RTEMS_EVENT_ANY | RTEMS_WAIT,
                                 RTEMS_NO_TIMEOUT,
                                 &events );

#ifdef DEBUG_NE2000
    printk ("ne_tx_daemon\n");
#endif

    /* This daemon handles both uploading data onto the device and
       telling the device to transmit data which has been uploaded.
       These are separate tasks, because while the device is
       transmitting one buffer we will upload another.  */

    /* Don't let the device interrupt us now.  */
    (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_IMR, 0 );

    while (1) {
      struct mbuf *m;

      /* If the device is not transmitting a packet, and we have
         uploaded a packet, tell the device to transmit it.  */
      if (! dp->transmitting && dp->inuse > 0) {
        dp->transmitting = 1;
        ns8390_transmit (dp);
      }

      /* If we don't have any more buffers to send, quit now.  */
      if (ifp->if_snd.ifq_head == NULL) {
        ifp->if_flags &= ~IFF_OACTIVE;
        break;
      }

      /* Allocate a buffer to load data into.  If there are none
         available, quit until a buffer has been transmitted.  */
      if (dp->inuse >= NS8390_TX_BUFS)
        break;

      ++dp->inuse;

      IF_DEQUEUE (&ifp->if_snd, m);
      if (m == NULL)
        panic ("ns8390_tx_daemon");

      ns8390_loadpacket (dp, m);

      /* Check the device status.  It may have finished transmitting
         the last packet.  */
      ne_check_status (dp);
    }

    /* Reenable device interrupts.  */
    (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_IMR, NSD8390_INTERRUPTS );
  }
}


/* trigger the send of a packet */

static void
ns8390_start (struct ifnet *ifp)
{
  struct ns8390_device  *dp = ifp->if_softc;

  if (debug_level >= DEBUG_STD)
    printf( "ns8390_start() - start\n" );

  rtems_event_send (dp->txDaemonTid, START_TRANSMIT_EVENT);
  ifp->if_flags |= IFF_OACTIVE;

  if (debug_level >= DEBUG_STD)
    printf( "ns8390_start() - end\n" );
}





/* send a packet */
static void
ns8390_sendpacket (struct ifnet *ifp, struct mbuf *m)
{
  struct ns8390_device  *dp = ifp->if_softc;
  unsigned8             status;
  unsigned int          len;

  unsigned32            ns8390_port;

  setRegister_f         setReg;
  getRegister_f         getReg;
  getData_f             getData;

  struct mbuf           *n;

  if (debug_level >= DEBUG_STD)
    printf("ns8390_sendpacket() - start\n");

  /* setup the IO functions */
  ns8390_port = dp->port;
  setReg    = dp->setRegister;
  getReg    = dp->getRegister;
  getData   = dp->getData;

  /* get the host driver to actually send the data into the packet buffer */
  dp->block_output( dp, &len, m, dp->tx_start_page );

  if (debug_level >= DEBUG_STD) 
    printf( "A %i byte packet has been loaded.\n", len );

  /* Free the mbufs */
  do
  {
    MFREE( m, n ); 
    m = n;
  } while ( m != NULL );

  /* upsize runt packets */

  if (len < ET_MINLEN) len = ET_MINLEN;


  /* Make sure we're on PAGE0 and abort any transfer */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB );

  /* Check we're not actually tranmitting */
  status = (*getReg)( ns8390_port, NS8390_PAGE0_READ_CR );
  if (status & NS8390_CR_TXP)
  {
    printf( "FreeBSD (%s) : Transmit attempted when the transmitter is busy.\n", ifp->if_name );
    return;
  }
  

  /* Set the packet length */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_TBCR0, len & 0xFF );
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_TBCR1, (len >> 8) & 0xFF );


  /* Where is the packet to transmit in internal memory */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_TPSR, dp->tx_start_page );

  /* Thunderpackets are Go! */
  (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_RD_AB | NS8390_CR_STA | NS8390_CR_TXP );

  if (debug_level >= DEBUG_STD)
    printf( "ns8390_sendpacket() - end\n" );
  
  /*
   * Show that we've finished with the packet - reset the ISR 
   */
}

static void
ns8390_rx_daemon (void *arg)
{
  struct ns8390_device  *dp  = (struct ns8390_device *)arg;
  struct ifnet      *ifp = &dp->arpcom.ac_if;
  struct mbuf       *bp;
    unsigned8       rxing_page;
    unsigned8       this_frame;
    unsigned8       next_frame;
    unsigned16        current_offset;
  int           pkt_len;

    int           num_rx_pages = dp->stop_page - dp->rx_start_page;


  struct ns8390_pkt_hdr rx_frame;

  unsigned32              ns8390_port;

  setRegister_f           setReg;
  getRegister_f           getReg;
  getData_f       getData;

  rtems_event_set     events;
  struct ether_header   *eh;


  if (debug_level >= DEBUG_STD)
    printf("Some silly fool is trying to call ns8390_rx()\n");

  /* setup the IO functions */
  ns8390_port = dp->port;
  setReg    = dp->setRegister;
  getReg    = dp->getRegister;
  getData   = dp->getData;


  /*
   * Allocate space for incoming packets and start reception
   */

  MGETHDR (bp, M_WAIT, MT_DATA);
  MCLGET (bp, M_WAIT);
  bp->m_pkthdr.rcvif = ifp;


  for ( ;; )
  {


      /* Get the rx page (incoming packet pointer). */
    (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE1 | NS8390_CR_RD_AB );
    rxing_page = (*getReg)( ns8390_port, NS8390_PAGE1_CURR );
    (*setReg)( ns8390_port, NS8390_PAGE1_CR, NS8390_CR_PAGE0 | NS8390_CR_RD_AB );


    /* Remove one frame from the ring.  Boundary is always a page behind. */
    this_frame = (*getReg)( ns8390_port, NS8390_PAGE0_READ_BNRY ) + 1;
    if (this_frame >= dp->stop_page)
      this_frame = dp->rx_start_page;

    /* Someday we'll omit the previous, if we never get this message.
       (There is at least one clone claimed to have a problem.)  */
    if (this_frame != dp->current_page)
      printf("%s: mismatched read page pointers %2x vs %2x.\n",
           ifp->if_name, this_frame, dp->current_page);

    if (this_frame == rxing_page) /* Read all the frames? */
        {
          /* hang about waiting for the next rx interrupt */
      if (debug_level >= DEBUG_STD)
        printf( "I'll hang about waiting for an INTERRUPT_EVENT...\n");

      rtems_bsdnet_event_receive (INTERRUPT_EVENT,
            RTEMS_WAIT|RTEMS_EVENT_ANY,
            RTEMS_NO_TIMEOUT,
            &events);

      if (debug_level >= DEBUG_STD)
        printf( "I got an INTERRUPT_EVENT!\n" );
        }
        else
        {
      /* recieve a packet */

            current_offset = this_frame << 8;

      /* suck the header from the chip */
      dp->get_ns8390_hdr( dp, &rx_frame, this_frame );


      pkt_len = rx_frame.count - sizeof(struct ns8390_pkt_hdr);

      next_frame = this_frame + 1 + ((pkt_len+4)>>8);

      /* This code carries out a "bogosity" check as recommended in the */
      /* 3c509 manual. Lifted straight from Linux             */
      if ( rx_frame.next != next_frame                    &&
         rx_frame.next != next_frame + 1                &&
         rx_frame.next != next_frame - num_rx_pages     &&
         rx_frame.next != next_frame + 1 - num_rx_pages    )
      {
        /* correct the current_page pointer */
        dp->current_page = rxing_page;

        /* write the updated boundary to the chip */
        (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_BNRY, dp->current_page - 1 );

        /* update the statistics */
        dp->eth_stats.rx_errors++;

        /* skip the rest of the packet recieve loop */
        continue;
      }
        

      if ( (pkt_len < 60) || (pkt_len > 1518) )
      {
        printf( "bogus packet size = %d\n", pkt_len);
        dp->eth_stats.rx_errors++;
      }
      else if ( (rx_frame.status & 0x0F) == NS8390_RSR_PRX )
      {

        dp->block_input( dp, pkt_len, bp, current_offset + sizeof(rx_frame) );


        bp->m_len = bp->m_pkthdr.len = pkt_len - sizeof(struct ether_header);
        eh = mtod (bp, struct ether_header *);
        bp->m_data += sizeof(struct ether_header);

        ether_input (ifp, eh, bp);

        
        dp->eth_stats.rx_packets++;

        MGETHDR (bp, M_WAIT, MT_DATA);
        MCLGET (bp, M_WAIT);
        bp->m_pkthdr.rcvif = ifp;

      }
      else
      {
        if ( rx_frame.status & NS8390_RSR_FO )
          dp->eth_stats.rx_fifo_errors++;

        printf( "bogus packet status = %#2x\n", rx_frame.status );
      }

      next_frame = rx_frame.next;


      /* Check to see that we haven't written over the end of the buffer! */
      if (next_frame >= dp->stop_page)
      {
        printf( "frame inconsistency... we just tried to walk over the end chaps!\n" );
        next_frame = dp->rx_start_page;
      }

      dp->current_page = next_frame;

      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_BNRY, next_frame - 1 );
    }


    /* go round again */
  }


  /* we never get here! */
}

static void
ns8390_stats (struct ns8390_device *dp)
{
  if (debug_level >= DEBUG_STD)
    printf("Some silly fool is trying to call ns8390_stats()\n");

  printf( "TX Packets         : %8d\n", dp->eth_stats.tx_packets );
  printf( "RX Packets         : %8d\n", dp->eth_stats.rx_packets );

  printf( "TX Errors          : %8d\n", dp->eth_stats.tx_errors );
  printf( "RX Errors          : %8d\n", dp->eth_stats.rx_errors );

  printf( "TX Dropped         : %8d\n", dp->eth_stats.tx_dropped );
  printf( "RX Dropped         : %8d\n", dp->eth_stats.rx_dropped );

  printf( "TX Interrupts      : %8d\n", dp->eth_stats.tx_interrupts );
  printf( "RX Interrupts      : %8d\n", dp->eth_stats.rx_interrupts );

  printf( "Collisions         : %8d\n", dp->eth_stats.collisions );
  
  printf( "RX Length Errors   : %8d\n", dp->eth_stats.rx_length_errors );
  printf( "RX Overflow Errors : %8d\n", dp->eth_stats.rx_over_errors );
  printf( "RX CRC Errors      : %8d\n", dp->eth_stats.rx_crc_errors );
  printf( "RX Frame Errors    : %8d\n", dp->eth_stats.rx_frame_errors );
  printf( "RX FIFO Errors     : %8d\n", dp->eth_stats.rx_fifo_errors );
  printf( "RX Missed Errors   : %8d\n", dp->eth_stats.rx_missed_errors );

  printf( "TX Aborted Errors  : %8d\n", dp->eth_stats.tx_aborted_errors );
  printf( "TX Carrier Errors  : %8d\n", dp->eth_stats.tx_carrier_errors );
  printf( "TX FIFO Errors     : %8d\n", dp->eth_stats.tx_fifo_errors );
  printf( "TX Heatbeat Errors : %8d\n", dp->eth_stats.tx_heartbeat_errors );
  printf( "TX Window Errors   : %8d\n", dp->eth_stats.tx_window_errors );

}


/*
 * Driver ioctl handler
 */
static int
ns8390_ioctl (struct ifnet *ifp, int command, caddr_t data)
{
  struct ns8390_device  *dp = ifp->if_softc;
  int           error = 0;


  if (debug_level >= DEBUG_STD)
    printf( "ns8390_ioctl() - start\n" );

  switch (command) {
  case SIOCGIFADDR:
  case SIOCSIFADDR:
    ether_ioctl (ifp, command, data);
    break;

  case SIOCSIFFLAGS:
    switch (ifp->if_flags & (IFF_UP | IFF_RUNNING)) {
    case IFF_RUNNING:
      ns8390_stop (dp);
      break;

    case IFF_UP:
      ns8390_init (dp);
      break;

    case IFF_UP | IFF_RUNNING:
      ns8390_stop (dp);
      ns8390_init (dp);
      break;

    default:
      break;
    }
    break;

  case SIO_RTEMS_SHOW_STATS:
    ns8390_stats (dp);
    break;
    
  /*
   * FIXME: All sorts of multicast commands need to be added here!
   */
  default:
    error = EINVAL;
    break;
  }

  if (debug_level >= DEBUG_STD)
    printf( "ns8390_ioctl() - end\n" );

  return error;
}


/*
 * Find a free driver
 */
struct ns8390_device *
ns8390_get_free_device( void )
{
  struct ifnet      *ifp;
  int           device_number;

  /* Print module version information */
  printf( version );

  for (device_number = 0 ; device_number < NUM_NS8390_DEVICES ; device_number++) 
  {
    ifp = &ns8390_devices[device_number].arpcom.ac_if;
    if (ifp->if_softc == NULL)
      break;
  }
  
  if (device_number >= NUM_NS8390_DEVICES) 
  {
    printf ("Too many NS8390 based drivers.\n");
    return NULL;
  }
  else
  {
    /* Zero out the device structure */
    memset( &ns8390_devices[device_number], 0, sizeof(struct ns8390_device) );

    if (debug_level >= DEBUG_STD)
      printf( "ns8390_get_free_device() - end\n" );
    return &ns8390_devices[device_number];
  }
}



/* External functions called by ns8390.c users */

int
ns8390_driver_attach ( struct rtems_bsdnet_ifconfig *config, struct ns8390_device *dp )
{
  struct ifnet  *ifp = &dp->arpcom.ac_if;
  int       device_number;
  int       mtu;


  if (debug_level >= DEBUG_STD)
    printf( "ns8390_driver_attach() - start\n" );


  /* re-find our device number */
  for (device_number = 0 ; device_number < NUM_NS8390_DEVICES ; device_number++) 
  {
    if (dp == &ns8390_devices[device_number])
      break;
  }


  /*
   * Process options
   */
  if (config->hardware_address) 
  {
    memcpy (dp->arpcom.ac_enaddr, config->hardware_address, ETHER_ADDR_LEN);
  }
  else 
  {
    dp->get_eth_addr( dp, dp->arpcom.ac_enaddr );
  }

  if (config->mtu)
    mtu = config->mtu;
  else
    mtu = ETHERMTU;


  dp->accept_broadcast = !config->ignore_broadcast;


  /*
   * Set up network interface values
   */
  ifp->if_softc  = dp;
  ifp->if_unit   = device_number + 1;
  ifp->if_name   = "ns";
  ifp->if_mtu    = mtu;
  ifp->if_init   = ns8390_init;
  ifp->if_ioctl  = ns8390_ioctl;
  ifp->if_start  = ns8390_start;
  ifp->if_output = ether_output;
  ifp->if_flags  = IFF_BROADCAST | IFF_SIMPLEX;
  if (ifp->if_snd.ifq_maxlen == 0)
    ifp->if_snd.ifq_maxlen = ifqmaxlen;

  /*
   * Attach the interface
   */
  if_attach (ifp);
  ether_ifattach (ifp);

  printf ("Ethernet address: %X%X%X%X%X%X\n", dp->arpcom.ac_enaddr[0], dp->arpcom.ac_enaddr[1], dp->arpcom.ac_enaddr[2], 
    dp->arpcom.ac_enaddr[3], dp->arpcom.ac_enaddr[4], dp->arpcom.ac_enaddr[5] );


  if (debug_level >= DEBUG_STD)
    printf("ns8390_driver_attach() - end\n");

  return 1;
}


rtems_isr
ns8390_interrupt_handler (rtems_vector_number v)
{
  struct ns8390_device  *dp = &ns8390_devices[0];
  unsigned8       status;

  unsigned32              ns8390_port;

  setRegister_f           setReg;
  getRegister_f           getReg;
  getData_f       getData;

  /* setup the IO functions */
  ns8390_port = dp->port;
  setReg    = dp->setRegister;
  getReg    = dp->getRegister;
  getData   = dp->getData;

  /* Store the interrupt status */
  status = (*getReg)( ns8390_port, NS8390_PAGE0_READ_ISR );

  /* abort if invalid isr value */
  if (status == 0)
    return;

  /* OK receive interrupt */
  if (status & NS8390_ISR_PRX)
  {
    /* Reset the ISR */
    (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_PRX );

    /* Signal the receive task so we can get the packet */
    dp->eth_stats.rx_interrupts++;
    rtems_event_send (dp->rxDaemonTid, INTERRUPT_EVENT);
  }

  /* OK transmit interrupt */
  if (status & NS8390_ISR_PTX)
  {
    (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_PTX );

    /* signal the transmit task and allow further packets to be transmitted */
    dp->eth_stats.tx_interrupts++;
    rtems_event_send (dp->txDaemonTid, INTERRUPT_EVENT);
  }

  /* receive error  - count */
  if (status & NS8390_ISR_RXE)
  {
    rtems_panic( "HELP HELP - NS8390 receive error\n" );
    (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_RXE );
    dp->eth_stats.rx_errors++;
  }

  /* transmit error - count */
  if (status & NS8390_ISR_TXE)
  {
    rtems_panic( "HELP HELP - NS8390 transmit error\n" );
    (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_TXE );
    dp->eth_stats.tx_errors++;
  }

  /* buffer overflow error - count and report */
  if (status & NS8390_ISR_OVW)
  {
    rtems_panic( "HELP HELP - NS8390 buffer overflow\n" );
    (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_OVW );
    dp->eth_stats.rx_over_errors++;

  }

  /* counter overflow error - report */
  if (status & NS8390_ISR_CNT)
  {
    rtems_panic( "HELP HELP - NS8390 counter overflow\n" );
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_CNT );
    dp->eth_stats.rx_frame_errors++;
  }

  /* remote dma complete - ignore */
  if (status & NS8390_ISR_RDC)
      (*setReg)( ns8390_port, NS8390_PAGE0_WRITE_ISR, NS8390_ISR_RDC );
}


/*
 * Initialize and start the device
 */
static void
ns8390_init (void *arg)
{
  struct ns8390_device  *dp = arg;
  struct ifnet      *ifp = &dp->arpcom.ac_if;

  if (dp->txDaemonTid == 0) {

    /*
     * Set up NS8390 hardware
     */
    ns8390_initialise(dp);

    /*
     * Start driver tasks
     */
    dp->txDaemonTid = rtems_bsdnet_newproc ("NStx", 4096, ns8390_tx_daemon, dp);
    dp->rxDaemonTid = rtems_bsdnet_newproc ("NSrx", 4096, ns8390_rx_daemon, dp);

  }

  /*
   * Set flags appropriately
   */
  /*if (ifp->if_flags & IFF_PROMISC)
    m360.scc1.psmr |= 0x200;
  else
    m360.scc1.psmr &= ~0x200; */

  /* we can set the NS8390 promisc mode here I guess */

  /*
   * Tell the world that we're running.
   */
  ifp->if_flags |= IFF_RUNNING;

  /*
   * Enable receiver and transmitter
   */
  ns8390_startup(dp);
}



/*
 * Initialize the ethernet hardware
 */
void
ns8390_initialise( struct ns8390_device *dp )
{
  unsigned8       *hwaddr;
  unsigned32              ne2000_port;

  setRegister_f           setReg;
  getRegister_f           getReg;
  getData_f       getData;


  if (debug_level >= DEBUG_STD)
    printf( "ns8390_initialise() - start\n" );

  /* setup the IO functions */
  ne2000_port = dp->port;
  setReg    = dp->setRegister;
  getReg    = dp->getRegister;
  getData   = dp->getData;
  
  /* Make sure we're on PAGE0, abort any transfer and stop the chip */
  (*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_STP | NS8390_CR_RD_AB );
      
  /* set the Data Configuration Register defaults */
    (*setReg)( ne2000_port, NS8390_PAGE0_WRITE_DCR, NS8390_DCR_BOS | NS8390_DCR_WTS | NS8390_DCR_FT_8B );

  /* zero the dma transfer size */
    (*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR0, 0 );
    (*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RBCR1, 0 );

  /* set the receive config to monitor */
    (*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RCR, NS8390_RCR_MON );

  /* set the transmit config to ENDEC module loopback */
    (*setReg)( ne2000_port, NS8390_PAGE0_WRITE_TCR, NS8390_TCR_LOOP_I0 );

  /* set the transmit start page */
    (*setReg)( ne2000_port, NS8390_PAGE0_WRITE_TPSR, dp->tx_start_page );

  /* zero the tx packet buffer sizes */
  dp->tx1 = 0;
  dp->tx2 = 0;

  /* set start and stop receive addresses */
    (*setReg)( ne2000_port, NS8390_PAGE0_WRITE_PSTART, dp->rx_start_page );
    (*setReg)( ne2000_port, NS8390_PAGE0_WRITE_PSTOP, dp->stop_page );

  /* set internal boundary buffer limit */
    (*setReg)( ne2000_port, NS8390_PAGE0_WRITE_BNRY, dp->stop_page - 1 );

  /* set the current packet pointer to the start of the region */
  dp->current_page = dp->rx_start_page;

  /* mask all the chip's interrupts */
    (*setReg)( ne2000_port, NS8390_PAGE0_WRITE_IMR, 0x00 );

  /* disable all interrupts */
    (*setReg)( ne2000_port, NS8390_PAGE0_WRITE_ISR, 0xFF );

  /* Go to PAGE1, abort any transfer and stop the chip */
  (*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE1 | NS8390_CR_STP | NS8390_CR_RD_AB );


  /* program supplied hardware address into the chip */
  hwaddr = dp->arpcom.ac_enaddr;
  (*setReg)( ne2000_port, NS8390_PAGE1_PAR0, hwaddr[0] );
  (*setReg)( ne2000_port, NS8390_PAGE1_PAR1, hwaddr[1] );
  (*setReg)( ne2000_port, NS8390_PAGE1_PAR2, hwaddr[2] );
  (*setReg)( ne2000_port, NS8390_PAGE1_PAR3, hwaddr[3] );
  (*setReg)( ne2000_port, NS8390_PAGE1_PAR4, hwaddr[4] );
  (*setReg)( ne2000_port, NS8390_PAGE1_PAR5, hwaddr[5] );


  /* program the multicast address  - to zero because we don't do multicast */
  (*setReg)( ne2000_port, NS8390_PAGE1_MAR0, 0 );
  (*setReg)( ne2000_port, NS8390_PAGE1_MAR1, 0 );
  (*setReg)( ne2000_port, NS8390_PAGE1_MAR2, 0 );
  (*setReg)( ne2000_port, NS8390_PAGE1_MAR3, 0 );
  (*setReg)( ne2000_port, NS8390_PAGE1_MAR4, 0 );
  (*setReg)( ne2000_port, NS8390_PAGE1_MAR5, 0 );
  (*setReg)( ne2000_port, NS8390_PAGE1_MAR6, 0 );
  (*setReg)( ne2000_port, NS8390_PAGE1_MAR7, 0 );

  /* set current page register */
  (*setReg)( ne2000_port, NS8390_PAGE1_CURR, dp->rx_start_page );

  /* Return to PAGE0, abort any transfer and stop the chip */
  (*setReg)( ne2000_port, NS8390_PAGE1_CR, NS8390_CR_PAGE0 | NS8390_CR_STP | NS8390_CR_RD_AB );


  if (debug_level >= DEBUG_STD)
    printf( "ns8390_initialise() - end\n" );
}


void
ns8390_startup( struct ns8390_device *dp )
{
  unsigned32              ne2000_port;

  setRegister_f           setReg;
  getRegister_f           getReg;


  if (debug_level >= DEBUG_STD)
    printf("ns8390_startup() - start\n");

  /* setup the IO functions */
  ne2000_port = dp->port;
  setReg    = dp->setRegister;
  getReg    = dp->getRegister;
  

  /* Final NS8390 configuration and startup */

  /* disable all interrupts */
    (*setReg)( ne2000_port, NS8390_PAGE0_WRITE_ISR, 0xFF );

  /* unmask all interrupts */
  (*setReg)( ne2000_port, NS8390_PAGE0_WRITE_IMR, NS8390_IMR_PRXE |
                          NS8390_IMR_PTXE |
                          NS8390_IMR_RXEE |
                          NS8390_IMR_TXEE |
                          NS8390_IMR_OVWE |
                          NS8390_IMR_CNTE   );

  /* Make sure we're on PAGE0, abort any transfer and startup the chip */
  (*setReg)( ne2000_port, NS8390_PAGE0_WRITE_CR, NS8390_CR_PAGE0 | NS8390_CR_STA | NS8390_CR_RD_AB );

  /* set the transmit config for normal operation */
  (*setReg)( ne2000_port, NS8390_PAGE0_WRITE_TCR, 0 );

  /* set the receive config for normal operation */
  (*setReg)( ne2000_port, NS8390_PAGE0_WRITE_RCR, NS8390_RCR_AB );

  if (debug_level >= DEBUG_STD)
    printf("ns8390_startup() - end\n");
}
